/**
 * Copyright (c) 2020 Bubble, Inc.  All rights reserved.
 * For personal (non-commercial) use, see license: https://getbubblenow.com/bubble-license/
 */
package bubble.service.backup;

import bubble.cloud.storage.StorageServiceDriver;
import bubble.dao.cloud.CloudServiceDAO;
import bubble.model.cloud.BubbleBackup;
import bubble.model.cloud.CloudCredentials;
import bubble.model.cloud.CloudService;
import bubble.model.cloud.NetworkKeys;
import bubble.server.BubbleConfiguration;
import lombok.Cleanup;
import lombok.Getter;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.IOUtils;
import org.cobbzilla.util.collection.NameAndValue;
import org.cobbzilla.util.io.TempDir;
import org.cobbzilla.wizard.cache.redis.RedisService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.io.*;
import java.util.Arrays;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import static bubble.ApiConstants.HOME_DIR;
import static bubble.model.cloud.NetworkKeys.PARAM_STORAGE;
import static bubble.model.cloud.NetworkKeys.PARAM_STORAGE_CREDENTIALS;
import static java.io.File.createTempFile;
import static java.util.concurrent.TimeUnit.HOURS;
import static java.util.concurrent.TimeUnit.MINUTES;
import static org.apache.commons.io.FileUtils.copyDirectory;
import static org.cobbzilla.util.daemon.ZillaRuntime.die;
import static org.cobbzilla.util.io.FileUtil.*;
import static org.cobbzilla.util.json.JsonUtil.json;
import static org.cobbzilla.util.security.CryptStream.BUFFER_SIZE;
import static org.cobbzilla.util.security.CryptoUtil.encrypt;
import static org.cobbzilla.wizard.cache.redis.RedisService.EX;

@Service @Slf4j
public class RestoreService {

    public static final File RESTORE_DIR = new File(HOME_DIR, "restore");

    // this is how long the RestoreService will allow a restore, after the API starts
    public static final long RESTORE_WINDOW = HOURS.toMillis(24);
    public static final long RESTORE_WINDOW_SECONDS = RESTORE_WINDOW/1000;

    // this is how long bubble_restore_monitor.sh will allow a restore after it starts
    // we add some time because, in the ansible setup, the script starts (in role bubble) before the
    // API is started (in role finalizer)
    public static final long RESTORE_MONITOR_SCRIPT_TIMEOUT_SECONDS = RESTORE_WINDOW_SECONDS + MINUTES.toSeconds(5);

    private static final long RESTORE_LOCK_TIMEOUT = MINUTES.toMillis(31);
    private static final long RESTORE_DEADLOCK_TIMEOUT = MINUTES.toMillis(30);
    private static final File RESTORE_MARKER_FILE = new File(HOME_DIR, ".restore");

    @Autowired private CloudServiceDAO cloudDAO;

    @Autowired private RedisService redis;
    @Getter(lazy=true) private final RedisService restoreKeys = redis.prefixNamespace(getClass().getSimpleName());

    @Autowired private BubbleConfiguration configuration;

    public void registerRestore(String restoreKey, NetworkKeys keys) {
        getRestoreKeys().set(restoreKey, json(keys), EX, RESTORE_WINDOW_SECONDS);
    }

    public boolean isValidRestoreKey(String restoreKey) { return getRestoreKeys().exists(restoreKey); }

    public boolean isSelfRestoreStarted() {
        // only one restore may be started, hence
        return getRestoreKeys().keys("*").size() > 0 || RESTORE_MARKER_FILE.exists();
    }

    public boolean restore(String restoreKey, BubbleBackup backup) {
        final String thisNodeUuid = configuration.getThisNode().getUuid();
        final String thisNetworkUuid = configuration.getThisNode().getNetwork();
        String lock = null;
        try {
            lock = getRestoreKeys().lock(thisNetworkUuid, RESTORE_LOCK_TIMEOUT, RESTORE_DEADLOCK_TIMEOUT);

            final String keyJson = getRestoreKeys().get(restoreKey);
            if (keyJson == null) {
                log.error("restore: restoreKey not found: " + restoreKey);
                return false;
            }
            final NetworkKeys networkKeys = json(keyJson, NetworkKeys.class);
            final String storageJson = NameAndValue.find(networkKeys.getKeys(), PARAM_STORAGE);
            final String credentialsJson = NameAndValue.find(networkKeys.getKeys(), PARAM_STORAGE_CREDENTIALS);
            if (storageJson == null || credentialsJson == null) {
                log.error("restore: storage/credentials not found in NetworkKeys");
                return false;
            }

            final String[] existingFiles = RESTORE_DIR.list();
            if (existingFiles != null && existingFiles.length > 0) {
                log.error("restore: files already exist in " + abs(RESTORE_DIR) + ", cannot restore");
                return false;
            }

            final CloudService storageService = json(storageJson, CloudService.class)
                    .setCredentials(json(credentialsJson, CloudCredentials.class));
            try {
                @Cleanup final TempDir temp = fetchBackupFiles(thisNodeUuid, storageService, backup);
                copyDirectory(temp, RESTORE_DIR);
                log.info("restore: notifying system to restore from backup at: "+abs(RESTORE_DIR));
                touch(RESTORE_MARKER_FILE);
                return true;
            } catch (IOException e) {
                log.error("restore: error downloading backup ", e);
                return false;
            }
        } finally {
            if (lock != null) {
                getRestoreKeys().unlock(thisNetworkUuid, lock);
            }
        }
    }

    /**
     * Build temporal ZIP with (specified) backup files for specified network/node.
     * @return Created tar file with backup.
     */
    @NonNull public File buildRestorePackage(@NonNull final String nodeUuid, @NonNull final BubbleBackup backup,
                                             @NonNull final String passphrase) {
        final var storageService = cloudDAO.findByUuid(configuration.getThisNetwork().getStorage());
        if (storageService == null) return die("buildRestorePackage: cannot find network storage for this network");

        try {
            @Cleanup final var temp = fetchBackupFiles(nodeUuid, storageService, backup);
            final var zipFile = zipAllFiles(temp, createTempFile("backup-", ".zip"));
            final var encryptedFile = createTempFile("backup-", "");

            // encrypt file:
            @Cleanup final var fis = new FileInputStream(zipFile);
            @Cleanup final var fos = new FileOutputStream(encryptedFile);
            fos.write(encrypt(fis, passphrase));
            fos.flush();
            zipFile.delete();

            return encryptedFile;
        } catch (Exception e) {
            return die("buildRestorePackage: error downloading backup files ", e);
        }
    }

    @NonNull private static File zipAllFiles(@NonNull final TempDir src, @NonNull final File dest) {
        try {
            @Cleanup FileOutputStream fos = new FileOutputStream(dest);
            @Cleanup ZipOutputStream zos = new ZipOutputStream(fos);

            final var srcFiles = src.listFiles();
            for (final var srcFile : srcFiles) {
                zipFile(srcFile, srcFile.getName(), zos);
            }

            return dest;
        } catch (IOException e) {
            return die("zipAllFiles: error zipping folder " + abs(src));
        }
    }

    /**
     * Originally taken from https://www.baeldung.com/java-compress-and-uncompress#zip_directory
     */
    private static void zipFile(@NonNull final File fileToZip, @NonNull final String relativePath,
                                @NonNull final ZipOutputStream zipOut)
            throws IOException {
        if (fileToZip.isDirectory()) {
            final var path = relativePath + (relativePath.endsWith("/") ? "" : "/");

            zipOut.putNextEntry(new ZipEntry(path));
            zipOut.closeEntry();

            final var children = fileToZip.listFiles();
            for (final var childFile : children) {
                zipFile(childFile, path + childFile.getName(), zipOut);
            }
        } else {
            zipOut.putNextEntry(new ZipEntry(relativePath));

            var bytes = new byte[1024];
            int length;
            @Cleanup final var fis = new FileInputStream(fileToZip);
            while ((length = fis.read(bytes)) >= 0) {
                zipOut.write(bytes, 0, length);
            }
        }
    }

    @NonNull private TempDir fetchBackupFiles(@NonNull final String thisNodeUuid,
                                              @NonNull final CloudService storageService,
                                              @NonNull final BubbleBackup backup)
            throws IOException {
        final var path = StorageServiceDriver.getPath(backup.getPath());
        log.info("fetchBackupFiles: downloading backup from path=" + path);
        final var storageDriver = storageService.getStorageDriver(configuration);
        final var destTempDir = new TempDir();
        final var destTempDirAbs = abs(destTempDir) + "/";

        var listing = storageDriver.list(thisNodeUuid, path);
        while (true) {
            Arrays.stream(listing.getKeys())
                  .parallel()
                  .forEach(k -> {
                      final var logMsgPrefix = "fetchBackupFiles [" + k + "]: ";
                      log.info(logMsgPrefix + "downloading file");
                      final var file = new File(destTempDirAbs + k);
                      mkdirOrDie(file.getParentFile());
                      try {
                          @Cleanup final var in = storageDriver.read(thisNodeUuid, k);
                          try (var out = new BufferedOutputStream(new FileOutputStream(file), BUFFER_SIZE)) {
                              IOUtils.copyLarge(in, out);
                          }
                          log.info(logMsgPrefix + "successfully downloaded file");
                      } catch (Exception e) {
                          die(logMsgPrefix + "error downloading file", e);
                      }
                  });

            if (!listing.isTruncated()) break;
            listing = storageDriver.listNext(thisNodeUuid, listing.getListingId());
        }
        log.info("fetchBackupFiles: full download successful");
        return destTempDir;
    }
}
